// This may look like C code, but it is really -*- C++ -*-
/*
 *************************************************************************
 *
 *		     An XVT window to display a gray-scale image
 *
 * The present class is based on the XVTWindow to display a gray-scale image
 * as well as possible. Note that it might not be possible to display all
 * shades of gray the image has. So, the histogram equalization and gray-shade
 * subsampling may be necessary. Also, the class takes care of the situation
 * when the size of the window isn't equal the dimension of the image (either
 * the window was originally that way, or after resizing by the user). Note,
 * the dimension of the window doesn't have to be an integer multiple of the
 * dimension of the image, any ratio would do.
 *
 * Note the class XVTImageWindow inherits from IMAGE: this is because the
 * image is displayed as long as it physically exists. However, if you changed
 * the IMAGE after an XVTImageWindow object was created, be sure to call
 * refresh() to change the image appearance on the screen.
 *
 * $Id: image_win.h,v 2.1 1994/09/06 14:35:26 oleg Exp oleg $
 *
 *************************************************************************
 */

#ifndef __GNUC__
#pragma once
#else
#pragma interface
#endif

#ifndef _image_win_h
#define _image_win_h 1

#ifdef max
#undef max
#undef min
#endif
#include "image.h"

#include "Window.h"

				// It's just an index in the CLUT on Mac and
				// under UNIX
typedef unsigned char GRAYSHADE;

class CustomCLUT;

class XVTColor			// We use the 8-bit RGB color model
{
  friend class CustomCLUT;

  unsigned char red, green, blue;

public:
  XVTColor(const char _red, const char _green, const char _blue)
    : red(_red), green(_green), blue(_blue) {}
  XVTColor(const char * rgb_str);	// Parse string #rgb or #rrggbb
  XVTColor(const COLOR xvt_color)	// Convert Color to XVTColor
  	{ red = xvt_color >> 16; green = xvt_color >> 8; blue = xvt_color; }
  ~XVTColor(void) {}
  operator COLOR (void) const	// Return Color as used in XVT
  	{ return ((unsigned long)red << 16) | 
	         ((unsigned long)green<<8) | blue; }
};

				// Color Look-up table with reserved entries
				// for displaying gray-scale images as well
				// as possible
class CustomCLUT
{
  enum { No_reserved_gray_shades = 128, Total_no_grays = 256 };
  static GRAYSHADE table[Total_no_grays];
public:
  CustomCLUT(void);
  ~CustomCLUT(void) {}
  const GRAYSHADE grayshade_of(const GRAY pixel)
  	{ assert(pixel < Total_no_grays); return table[pixel]; }
  void set_palette(const WINDOW win);

				// Direct manipulation of the CLUT entries
  XVTColor get_entry(const GRAYSHADE index) const;
  void set_entry(const GRAYSHADE index, const XVTColor new_color);
};

				// Simplified pixmap class
class PixelMap: public CustomCLUT
{
public:
  short width, height;
  unsigned char * pixmap;
	
  PixelMap(void)	{ width = height = 0; pixmap = 0; }
  ~PixelMap(void) 	{ if( pixmap != 0 ) delete pixmap; }
  operator char * (void) const	{ return (char *)pixmap; }
					// Convert an image to pixmap with
					// possible resizing to fit the window
  void set_up(const XVTRect& dest_rect, const IMAGE& image);
};


class XVTImageWindow : public XVTWindow, public IMAGE
{
  PixelMap fitted_pixmap;
  
  virtual void draw(void);		// It is this function that draws smth

protected:
  virtual void handler(EVENT event);	// More specific handler
  virtual BOOLEAN initialize(void);	// Window is constructed at this moment

public:
  XVTImageWindow(const short resource_id, const IMAGE& image,
		 const char * title,const WINDOW parent_win=SCREEN_WIN);
  XVTImageWindow(const short resource_id, const short nrows, const short ncols,
		 const char * title,const WINDOW parent_win=SCREEN_WIN);
  XVTImageWindow(const short resource_id, const char * file_name,
		 const char * title,const WINDOW parent_win=SCREEN_WIN);
  ~XVTImageWindow(void) {}

					// We need this because assignment
					// operators are not inherited
  IMAGE& operator = (const int val)
				{ IMAGE::operator = (val); return *this; }
	
					// Refresh the display when the image
					// has been changed
  void refresh(void)	     { fitted_pixmap.set_up((XVTWindow&)*this,*this); }

					// Update the picture on the screen
  void redraw(void) const
		{ ::invalidate_rect(q_window(),XVTRect((XVTWindow&)*this)); }
};

#endif
