// This may look like C code, but it is really -*- C++ -*-
/*
 *************************************************************************
 *
 *			Basic Window class for XVT
 *
 *
 *	   A generic simple window with controls/subwindows (if any)
 *    that can be dragged around on the screen and closed by clicking a
 *			      "go-away" button
 *
 * The class interfaces with the XVT Event system and performs only basic
 * window processing chores: watching creation/destroying of the window,
 * basic handling of window events and referring them to some
 * "specialist"-handlers. Some of them are just vanilla (like handler of
 * control items), some are smack virtual (like draw()): the user is supposed
 * to define/redefine them in the derived classes depending on 
 * needs/functionality of his particular window.
 *
 * The file also implements some operations on controls/subwindows like
 * setting/getting the title, enabling/disabling, finding selection in the
 * LISTBOX, handling "CANCEL" button, etc.
 *
 * $Id: Window.cc,v 2.2 1994/11/07 20:34:21 oleg Exp oleg $
 *
 ***********************************************************************
 */

#ifndef __GNUC__
#pragma implementation
#endif
#include "Window.h"

/*
 *----------------------------------------------------------------------
 *				Service functions
 */
				// Shifting a rectangle by the same amount
				// in X and Y
XVTRect& XVTRect::operator += (const int offset)
{ 
  top += offset; bottom += offset; left += offset; right += offset;
  return *this;
}

				// Give a STATIC string representation of
				// the rectangle
XVTRect::operator const char * (void) const
{
  char buffer[100];
  sprintf(buffer,"(%d,%d)-(%d,%d)",left,top,right,bottom);
  return buffer;
}

/*
 *----------------------------------------------------------------------
 *		   Primitive Window creators/destructors
 */

			// This is a protected constructor. Normally
			// objects XVTBasicWindow are created as a part
			// of a derived class
XVTBasicWindow::XVTBasicWindow(void)
{
  Not_canceled = FALSE;
  this_window = NULL_WIN;
}

			// Destroy the window if it hasn't been destroyed
			// already
XVTBasicWindow::~XVTBasicWindow(void) 
{
  if( this_window != 0 ) 
    close_window(this_window);
}

			// Close the window
void XVTBasicWindow::cancel(void)
{ 
  Not_canceled = FALSE;
  if( this_window != 0 )
    close_window(this_window);
}

/*
 *----------------------------------------------------------------------
 *		     Basic event dispatching/handling
 */

			// A universal event handler: does almost nothing
			// but watches creation and destruction of windows
			// (to tell when there are no windows to watch)
			// For any other tasks, it calls the handler of
			// a particular window
int XVTBasicWindow::no_opened_windows = 0;
#ifdef __SC__
				// Variable number of arguments is just to
				// force the C calling convention. Normally
				// member functions are called using the
				// pascal conventions in Symantec C++
long XVTBasicWindow::universal_handler(WINDOW win, EVENT *ep,...)
#else
long XVTBasicWindow::universal_handler(WINDOW win, EVENT *ep)
#endif
{
  XVTBasicWindow& the_window = *(XVTBasicWindow *)get_app_data(win);
  	
  if( ep->type == E_CREATE )
    no_opened_windows++, the_window.this_window = win;
  	  
  the_window.handler(*ep);

				// XVT would destroy the window after this
				// event
  if( ep->type == E_DESTROY )
    no_opened_windows--, the_window.this_window = 0;
  	
  return 0;
}

				// Serve the windows until all get closed 
void XVTBasicWindow::serve_to_death(void)
{
  while( no_opened_windows>0 )
    process_events();
}


			// Event handler for a particular window
			// Finally, a handler which really does smth
			// specific
void XVTBasicWindow::handler(EVENT event)
{
  switch(event.type)
  {
    case E_CREATE:
         if( !initialize() )
	   cancel();
	 break;
			
    case E_DESTROY:
	 break;
			
    case E_CLOSE:			// Request to close the window   
	 cancel();			// (by clicking the close box)
	 break;
			
    case E_CHAR:      // Only char not consumed by control and text-edit fields
	 break;

    case E_UPDATE:			// Dialogs are not getting these events
	 draw();			// Paint/repaint the window
	 break;
			
    case E_COMMAND:
	 break;
		
    case E_CONTROL:
	 short id = event.v.ctl.id;
	 if( id == DLG_CANCEL )
	 {
	   cancel();
	   break;
	 }
	 else if( !handle_ctl(id,event.v.ctl.ci) )
	   close_window(this_window);
	 break;

    default:			// Ignore the rest
	 break;
  }
}


			// Dummy 'initialize' routine, always returns TRUE,
			// meaning proceed with the window
BOOLEAN XVTBasicWindow::initialize(void)
{
  Not_canceled = TRUE;
  return TRUE;
}

			// Dummy control handler routine, always returns FALSE,
			// finish with the window
BOOLEAN XVTBasicWindow::handle_ctl(const short id, const CONTROL_INFO& ctl)
{
  return FALSE;
}

/*
 *------------------------------------------------------------------------
 * 	Handling window items (controls mostly or child subwindows)
 */

			// Return a window associated with a control item
const WINDOW XVTBasicWindow::q_window(const short id) const
{
  WINDOW win = get_ctl_window(this_window,id);
  assert(win != 0);
  return win;
}

			// Set a title of a control item (say, button, etc)
void XVTBasicWindow::set_title(const short id, const char * text)
{
  ::set_title(q_window(id),(char *)text);
}

const char * XVTBasicWindow::get_title(const short id) const
{
  static char buffer[256];
  return ::get_title(q_window(id),buffer,sizeof(buffer));
}

			// Enable/disable a window item
void XVTBasicWindow::enable(const short id, const BOOLEAN onoff)
{
  ::enable_window(q_window(id),onoff);
}

			// Show/hide a window item
void XVTBasicWindow::show(const short id, const BOOLEAN onoff)
{
  ::show_window(q_window(id),onoff);
}

			// Count how many items were selected in the
			// list control
int XVTBasicWindow::count_selections(const short id) const
{
  short n = ::win_list_count_sel(q_window(id));
  assert( n >= 0 );
  return n;
}

			// Add a string to a list control
void XVTBasicWindow::add_to_list_control(const short id, const char * str)
{
  assert( ::win_list_add(q_window(id),-1,(char *)str) );
}

			// Get the index of the 1st selected item, index starts
			// with 0
int XVTBasicWindow::get_selection(const short id) const
{
  short n = ::win_list_get_sel_index(q_window(id));
  assert( n >= 0 );
  return n;
}

			// Select an item of the list control from within
			// the program
			// onoff=FALSE resets the selection. Index
			// starts with 0
			// index = -1 for some controls refers to all items
			// of the list control
void XVTBasicWindow::set_selection(const short id, const short item_index,
			      const BOOLEAN onoff)
{
  assert( ::win_list_set_sel(q_window(id),item_index,onoff) );
}

			// Set the state of a check button
void XVTBasicWindow::set_checked(const short id,const BOOLEAN onoff)
{
  ::win_check_box(q_window(id),onoff);
}

			// Add a list to the list control
void XVTBasicWindow::add_list(const short id, const ListofStrings& list)
{
  assert( ::win_list_add(q_window(id),-1,(char *)list.list) );
}

			// Remove all items from the list control
void XVTBasicWindow::clear_list(const short id)
{
  assert( ::win_list_clear(q_window(id)) );
}

				// Indicate that a long operation is
				// in progress
void XVTBasicWindow::set_cursor_wait(const BOOLEAN onoff)
{
  ::set_cursor(q_window(),onoff ? CURSOR_WAIT : CURSOR_ARROW);
}

/*
 *----------------------------------------------------------------------
 *				Window creators
 *
 * Note, event processing may start *before* the create_* functions return!
 * So, if initialize() or event handlers, etc rely on some data that are
 * going to be initialized *after* create_* function, you're in trouble!
 * It's better to call create() function the last thing in the constructor,
 * or even not to call at all (call it separetely later) to allow derived
 * classes initialize properly!
 *
 */
			// Create a window of a specified size and title
			// Note, the upper left corner of 'rect' must
			// NOT be (0,0)!
void XVTWindow::create(XVTRect rect, const char * title)
{
  const int window_offset = 100;		// Cannot be 0 or small!
  create_window(W_DOC, rect+=window_offset/*XVT_MAX_WINDOW_RECT*/, 
		(char *)title, 0, SCREEN_WIN, 
		WSF_NO_MENUBAR | WSF_INVISIBLE | WSF_SIZE | WSF_CLOSE, 
		EM_ALL, (EVENT_HANDLER)q_universal_handler(),(long)this);
}

			// Create a child window laying over a parent window
void XVTWindow::create(const XVTBasicWindow& parent)
{
  create_window(W_NO_BORDER, XVTRect(parent),"", 0, parent.q_window(),
		WSF_INVISIBLE, 
		EM_ALL, (EVENT_HANDLER)q_universal_handler(),(long)this);
}

			// Create a window from the resource
			// Make sure that the window is declared initially
			// invisible in the resource file
void XVTWindow::create(const short resource_id,
			       const WINDOW parent_win)
{
  create_res_window(resource_id,parent_win,EM_ALL,
		    (EVENT_HANDLER)q_universal_handler(),(long)this);
}
