// This may look like C code, but it is really -*- C++ -*-
/*
 *************************************************************************
 *
 *		     A window to display a gray-scale image
 *
 * The present class is based on the XVTWindow to display a gray-scale image
 * as well as possible. The class performs the histogram equalization of the
 * image before its display. Moreover, the class performs resizing of the
 * image to exactly fit the window. Resizing is done in a dumb way, by
 * duplicating/dropping image pixels when appropriate. The resizing algo can
 * accomodate any ratio between the dimensions of the window and the image.
 *
 * Image display is done by moving an off-screen pixel map to the corresponding
 * part of the screen (where the image is supposed to be displayed). 
 * It is platform specific. At the moment, only X11 and Mac platforms are
 * supported. Also, a special class CustomCLUT takes care of reserving
 * certain cells in the LUT for gray shades (to avoid conflicts with other
 * windows/applications when displaying a full grayscale image).
 *
 * $Id: image_win.cc,v 2.4 1994/10/06 15:21:23 oleg Exp oleg $
 *
 *************************************************************************
 */

#ifndef __GNUC__
#pragma implementation
#endif
#include "image_win.h"

/*
 *----------------------------------------------------------------------
 *		Operations on the custom color look-up table
 */

GRAYSHADE CustomCLUT::table[CustomCLUT::Total_no_grays];

		// Reservation of gray shades in the native window system

#if (XVTWS >= X11WS)
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

		// Under X11, to save ourselves future trouble and to
		// guarantee we have enough (at least 128) private color
		// cells for our gray business, we allocate our own private
		// colormap and associate it with our window.
		// To be kind to other windows (say, to display window
		// borders, etc in their original colors), and to avoid
		// unnecessary color flushes when our colormap is installed,
		// we copy at least a quater of the "standard" colormap in
		// our private colormap.
static Colormap private_cmap;

CustomCLUT::CustomCLUT(void)
{
  static BOOLEAN already_set_up = FALSE;

  if( already_set_up )
    return;

  already_set_up = TRUE;

  Display * display = (Display *)get_value(SCREEN_WIN,ATTR_X_DISPLAY);
  assert( display != 0 );
  int x_screen = DefaultScreen(display);
  Colormap standard_cmap;

  assert( (standard_cmap = DefaultColormap(display,x_screen)) != 0 );
  int total_cmap_cells = DisplayCells(display,x_screen);
  assert( total_cmap_cells > 0 );
  if( total_cmap_cells < No_reserved_gray_shades+2 )
    xvt_fatal("Too few, %ld, colormap cells are available on your X server",
	      total_cmap_cells);

  Window x_window = (Window)get_value(TASK_WIN,ATTR_NATIVE_WINDOW);
  assert( (private_cmap=XCreateColormap(display,
				    x_window,
				    /*DefaultRootWindow(display),*/
				    DefaultVisual(display,x_screen),
				    AllocNone)) != 0 );
//  xvt_note("Standard cmap 0x%lx, private cmap 0x%lx, task window 0x%lx",
//	   standard_cmap,private_cmap,x_window);

  {				// Copy 1/4 of standard_cmap into private_cmap
				// and make them shared
    int no_cells_to_copy = total_cmap_cells/4;
    assert( no_cells_to_copy > 2 );
    XColor * colors = new XColor[no_cells_to_copy];
    assert( colors != 0 );

    register int i;
    for(i=0; i<no_cells_to_copy; i++)
      colors[i].pixel = i, colors[i].flags = DoRed | DoGreen | DoBlue;

    XQueryColors(display,standard_cmap,colors,no_cells_to_copy);

    for(i=0; i<no_cells_to_copy; i++)
      assert( XAllocColor(display,private_cmap,&colors[i]) );

    delete colors;
  }

				// Allocate grayshades in the rest of the
				// color map.
				// Make them private, though, so we can
				// change them later
  register int i,j;
  unsigned long allocated_pixels[No_reserved_gray_shades];

  assert( XAllocColorCells(display,private_cmap,TRUE,(void *)0,0,
			   allocated_pixels,No_reserved_gray_shades) );
  for(i=0,j=0; i<No_reserved_gray_shades; i++)
  {
    int gray_shade = (long)i*Total_no_grays/No_reserved_gray_shades;
    XColor xcolor;
    xcolor.red = xcolor.green = xcolor.blue = (gray_shade << 8) | gray_shade;
    xcolor.flags = DoRed | DoGreen | DoBlue;
    xcolor.pixel = allocated_pixels[i];
    XStoreColor(display,private_cmap,&xcolor);
    while( j<(long)(i+1)*Total_no_grays/No_reserved_gray_shades )
      table[j++] = xcolor.pixel;
  }
  assert( j == Total_no_grays );

  xvt_note("reserved gray shades");
}

			// Associate our colormap with the window
			// Note, XVT window turns out to be implemented as a
			// series of at least 3 nested windows.
			// (Maybe it refers only to W_DOC windows).
			// The topmost window contain the menu bar, the border,
			// etc. And it's this window which should contain
			// the private cmap (because it's only when the
			// colormap gets installed by the win manager).
			// This window is the only one in the hierarchy that
			// has a name, btw. BTW, there is some other window
			// between this, named, window and the root window.
			// After the colormap is set up for the parent, we
			// set it up for all existing kids (it takes care of
			// controls): the situation with controls arises when
			// the image window is a child of a bigger window
			// which has other subwindows (controls) beside the
			// image window

				// Associate private_cmap with an x_window
				// and its kids (recursively)
static void set_kids_cmap(Display * display, Window x_window)
{
  XSetWindowColormap(display,x_window,private_cmap);

  Window root;			// Find kids for a window
  Window parent;
  Window *kids;
  unsigned int nkids;
  assert( XQueryTree(display,x_window,&root,&parent, &kids, &nkids) );
  register int i;
  for(i=0; i<nkids; i++)
    set_kids_cmap(display,kids[i]);
      
  XFree((char *)kids);
}

void CustomCLUT::set_palette(const WINDOW win)
{
  assert( win != NULL_WIN );
  Display * display = (Display *)get_value(SCREEN_WIN,ATTR_X_DISPLAY);
  assert( display != 0 );
  Window x_our_window = (Window)get_value(win,ATTR_NATIVE_WINDOW);

					// Looking from x_our_window up
					// for a window that has a name
  for(;;)
  {						// Associate CMAP anyway
//    XSetWindowColormap(display,x_our_window,private_cmap);
//    xvt_note("=== installing cmap 0x%x in window 0x%x",private_cmap,
//	     x_our_window);
    char * win_name;
    XFetchName(display,x_our_window,&win_name);
    if( win_name != 0 )
    {					// Found a window with a name,
      XFree(win_name);			// it is the top of our hierarchy
      break;
    }
    
    Window root;			// Find a parent for a window
    Window parent;
    Window *kids;
    unsigned int nkids;
    assert( XQueryTree(display,x_our_window,&root,&parent, &kids, &nkids) );
    XFree((char *)kids);
    if( parent == root )
      _error("no parent");		// reached the top of hierarchy
    x_our_window = parent;		// otherwise, continue setting CMAP
  }
  set_kids_cmap(display,x_our_window);
}

				// Query the RGB settings of a specific
				// color cell (CLUT entry with a given index)
XVTColor CustomCLUT::get_entry(const GRAYSHADE index) const
{
  Display * display = (Display *)get_value(SCREEN_WIN,ATTR_X_DISPLAY);
  assert( display != 0 );
  XColor xcolor;
  xcolor.pixel = index;
  XQueryColor(display,private_cmap,&xcolor);
					// xcolor.red etc are 16-bit values,
					// but usually at most 8 high-order
					// bits are important
  return XVTColor(xcolor.red>>8, xcolor.green>>8, xcolor.blue>>8);
}

				// Explicitly set up a colorcell, that is,
				// assign an RGB color to a CLUT entry.
				// It has an immediate effect of changing
				// corresponding pixels on the screen
void CustomCLUT::set_entry(const GRAYSHADE index, const XVTColor new_color)
{
  Display * display = (Display *)get_value(SCREEN_WIN,ATTR_X_DISPLAY);
  assert( display != 0 );
  XColor xcolor;
  xcolor.pixel = index;				// Make a 16-bit color value
  xcolor.red   = ( new_color.red << 8 ) | new_color.red;  // from the 8-bit
  xcolor.green = ( new_color.green << 8 ) | new_color.green;
  xcolor.blue  = ( new_color.blue << 8 ) | new_color.blue;
  xcolor.flags = DoRed | DoGreen | DoBlue;
  XStoreColor(display,private_cmap,&xcolor);
}

#elif (XVTWS==MACWS)
CustomCLUT::CustomCLUT(void)
{
  static Boolean already_set_up = FALSE;

  if( already_set_up )
    return;

  already_set_up = TRUE;
  register int i;
  int degradation_factor = Total_no_grays/No_reserved_gray_shades;
  assert( degradation_factor*No_reserved_gray_shades == Total_no_grays);

  for(i=0; i<Total_no_grays; )
  {
    int gray_shade = i;
    register int j;
    for(j=0; j<degradation_factor; j++)
      table[i++] = ~gray_shade;			// Perform the inversion
    assert( i <= Total_no_grays );
  }
}

				// Associate our colormap with the window
void CustomCLUT::set_palette(const WINDOW win)
{
}
#endif


/*
 *----------------------------------------------------------------------
 *			Operations on a pixmap
 */

				// Convert an image to pixmap with possible
				// resizing to exactly fit the window
				// Squeezing is done by subsampling (dropping
				// some image pixels), while stretching is done
				// by repeating pixels/scanlines
				// At the same time, we convert image pixels
				// according to the CLUT we've reserved
				// Note, dest_rect is a rectangle (in windows
				// coord) where we're drawing
void PixelMap::set_up(const XVTRect& dest_rect, const IMAGE& image)
{
					// Make the width even
  width  = ( dest_rect.q_width() + 1 ) & (~1);
  height = dest_rect.q_height();
  assert( width > 0 && height > 0 );

  if( pixmap != 0 )
    delete pixmap;
  assert( (pixmap = (unsigned char *)malloc((long)width*height)) != 0 );

					// Figure out how much to
					// sqeeze/stretch
  float ratio_h = (float)width/image.q_ncols();
  float ratio_v = (float)height/image.q_nrows();

  float curr_v = 0, curr_h;		// Precise curr. loc. in window coord
  int dest_y = 0, dest_x;		// approx (rounded-off) coordinates
					// Scanline ptr for fitted_pixmap
  register unsigned char * npr = (unsigned char *)(char *)pixmap;
  register int i,j;

 					// Note, dest_x,dest_y is incremented
					// by (inc_v,inc_h), but curr_v is
					// incremented by ratio_v, etc.
					// Draw horizontally (row-by-row)
  for(i=0; i<image.q_nrows(); i++)
  {
    curr_v += ratio_v;

    int inc_v = (int)curr_v - dest_y;
    if( inc_v == 0 )			// inc_v = 0 means this row of 
      continue;				// orig_image is to be skipped
    dest_y = (int)curr_v;

    curr_h = 0;
    dest_x = 0;
    unsigned char * npp;		// Squeeze/stretch pixels in a scanline
    for(npp=npr,j=0; j<image.q_ncols(); j++)
    {
      unsigned int pixel = grayshade_of(image(i,j));
      curr_h += ratio_h;

      int inc_h = (int)curr_h - dest_x;
      dest_x = (int)curr_h;
      while( inc_h-- > 0 )
	*npp++ = pixel;
    }
    assert( npp == npr+width || npp == npr+width-1 );
					// Stretching vertically involves
					// duplicating scanlines
    while( --inc_v > 0 )
      memcpy((void *)(npr+width),(void *)npr,width),
      npr += width;
    npr += width;
  }
  assert( npr == pixmap + (long)width*height ||
	  npr == pixmap + (long)width*(height-1) );
}

/*
 *----------------------------------------------------------------------
 *			Drawing an image in a window
 */

XVTImageWindow::XVTImageWindow(const short resource_id, const IMAGE& image, 
			       const char * title,const WINDOW parent_win)
	: IMAGE(image)
{
  (IMAGE&)*this = image;
  create(resource_id,parent_win);
}

XVTImageWindow::XVTImageWindow(const short resource_id, const short nrows,
			       const short ncols, const char * title,
			       const WINDOW parent_win)
	: IMAGE(nrows,ncols,8)
{
  create(resource_id,parent_win);
}

XVTImageWindow::XVTImageWindow(const short resource_id, const char * file_name,
			       const char * title,const WINDOW parent_win)
	: IMAGE(file_name)
{
  equalize(128);
  create(resource_id,parent_win);
}

				// Window is constructed at this moment
				// Note, we don't do refresh here, as
				// E_SIZE event is sure to follow
BOOLEAN XVTImageWindow::initialize(void)
{
  fitted_pixmap.set_palette(q_window());
  return TRUE;
}
				// It handles in particular only the size event
				// The rest of events are handled by a generic
				// window event handler
void XVTImageWindow::handler(EVENT event)
{
  if( event.type == E_SIZE )
    refresh(),
    redraw();

  XVTWindow::handler(event);
}


				// Actual drawing - moving the picture from the
				// offscreen grafworld to the onscreen one
#if (XVTWS >= X11WS)
void XVTImageWindow::draw(void)
{
  const WINDOW xvt_win = q_window();
  Display * display = (Display *)get_value(xvt_win,ATTR_X_DISPLAY);
  GC gc = (GC)get_value(xvt_win,ATTR_NATIVE_GRAPHIC_CONTEXT);
  Window x_window = (Window)get_value(xvt_win,ATTR_NATIVE_WINDOW);

  XWindowAttributes x_win_attr;
  assert( XGetWindowAttributes(display,x_window,&x_win_attr) );

  XImage * x_image = XCreateImage(display,x_win_attr.visual,8,ZPixmap,0,
				  (char *)fitted_pixmap,fitted_pixmap.width,
				  fitted_pixmap.height, 8,0);
  assert( x_image != 0 );
  XPutImage(display,x_window,gc,x_image, 0,0, 0,0, 
	    fitted_pixmap.width, fitted_pixmap.height);
  XFree((char *)x_image);		// frees only the XImage structure,
					// fitted_pixmap ptr remains intact
}

#elif (XVTWS==MACWS)
void XVTImageWindow::draw(void)
{
  GrafPtr old_port;
  GrafPtr our_port = (GrafPtr)get_value(q_window(),ATTR_NATIVE_WINDOW);
  assert(our_port != nil);
  GetPort(&old_port);
  SetPort(our_port);

				// Figuring out where to draw
				// Note that our_port is the parent window's
				// grafport, only get_outer_rect() gives us the
				// rectangle for our window in the parent
				// grafport's coordinates.
				// But this rectangle includes the window
				// border, which we should account for
  RCT outer_rect; assert( get_outer_rect(q_window(),&outer_rect) );
  WIN_TYPE win_type = get_window_type(q_window());
  int offset_x, offset_y;
  switch( win_type )
  {
    case W_DOC:
         offset_x = get_value(NULL,ATTR_DOCFRAME_WIDTH)/2;
	 offset_y = get_value(NULL,ATTR_DOCFRAME_HEIGHT)/2;
	 break;

    case W_PLAIN:
	 offset_x = get_value(NULL,ATTR_FRAME_WIDTH)/2;
	 offset_y = get_value(NULL,ATTR_FRAME_HEIGHT)/2;
	 break;

    case W_NO_BORDER:
	 offset_x = offset_y = 0;
	 break;
	  	
    default:
	 _error("Don't know window type %ld",win_type);
  }

				// These are the coordinates of the upper
				// left corner of the image window in our_port
				// coordinates
  int orig_x = outer_rect.left + offset_x;
  int orig_y = outer_rect.top + offset_y;
  //Rect portrect = (*(((CGrafPort *)our_port)->portPixMap))->bounds;
  //_error("Portrect (%ld,%ld)-(%ld,%ld)",portrect.left,portrect.top,
  //		portrect.right, portrect.bottom);

  PixMapHandle pixmap_handle;
  assert( (pixmap_handle = NewPixMap()) != nil );
  
				// Create an off-screen pixmap that points to
				// our 'fitted_pixmap' pixel matrix
  (**pixmap_handle).baseAddr = (char *)fitted_pixmap;
  assert( (fitted_pixmap.width & 1) == 0 );		// Must be even
  (**pixmap_handle).rowBytes = 0x8000 | fitted_pixmap.width;
  (**pixmap_handle).bounds.top = 0;
  (**pixmap_handle).bounds.left = 0;
  (**pixmap_handle).bounds.right = fitted_pixmap.width;
  (**pixmap_handle).bounds.bottom = fitted_pixmap.height;
  (**pixmap_handle).packType = 0;			// No pixel packing
  (**pixmap_handle).packSize = 0;
  (**pixmap_handle).pixelType = 0;	// Chunky pixels (not color planes)
  (**pixmap_handle).pixelSize = 8;
  (**pixmap_handle).cmpCount = 1;
  (**pixmap_handle).cmpSize = 8;
  (**pixmap_handle).planeBytes = 0;
	
  DisposCTable((**pixmap_handle).pmTable);// Assign the full-gray color table
  assert( ((**pixmap_handle).pmTable = GetCTable(40)) != nil );
	
				// Where is our window to display the image in
  Rect where_rect = (**pixmap_handle).bounds;
  where_rect.left   += orig_x;
  where_rect.top    += orig_y;
  where_rect.bottom += orig_y;
  where_rect.right  += orig_x;
  
				// Draw the pixmap on the screen
  HLock((char **)pixmap_handle);
  CopyBits((BitMap *)*pixmap_handle,
	   (BitMap *)*(((CGrafPort *)our_port)->portPixMap), 
	   &(**pixmap_handle).bounds, &where_rect, /*ditherCopy*/srcCopy, nil);
  HUnlock((char **)pixmap_handle);
  (**pixmap_handle).baseAddr = nil;
  DisposPixMap(pixmap_handle);
    
  SetPort(old_port);
}
#endif
