// This may look like C code, but it is really -*- C++ -*-
/*
 *************************************************************************
 *
 *				Rubber rectangle
 * 				  Implementation
 *
 * As was mentioned in rubber_rect.h, the rubber rectangle is drawn on
 * a specially created transparent window, which is a child window of
 * the XVTImageWindow. Note, during the life of the RubberRectangle
 * object, that child window (where the rectangle is displayed) can be
 * opened and closed several times. Also, be sure that the RubberRectangle
 * objects lives no longer than XVTImageWindow!
 *
 * Under X Windows, the system always clears the window once it was
 * created (and becomes visible) using the background pixel/background
 * pattern. To make window transparent (that's what we need) the background
 * pattern should be set to None. It can be done when the window is created
 * by XCreateWindow. Unfortunately, XVT must be using XCreateSimpleWindow
 * which doesn't let one specify the background pattern from the outset.
 * Moreover, drawing tools do not affect the default window background
 * color/pattern. The bottom line is, XVT sucks once again in handling
 * X windows, and we need to kludge using X calls (change attributes of
 * an X window while it stays unvisible (it was created that way)).
 *
 * $Id: rubber_rect.cc,v 1.1 1994/11/10 20:14:52 oleg Exp oleg $
 *
 ***********************************************************************
 */

#ifdef __GNUC__
#pragma implementation
#endif

#include "rubber_rect.h"
#include <std.h>

#if (XVTWS >= X11WS)
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif

/*
 *------------------------------------------------------------------------
 *			Constructing the Rubber Rectangle
 */

				// This construction does some intialization
				// but it does _not_ create a window nor
				// shows the rectangle
RubberRectangle::RubberRectangle(const XVTImageWindow& _parent)
	: parent(_parent), rect((XVTBasicWindow&)_parent)
{
  parent.is_valid();
  init_rectangle();
}


				// Make an initial rectangle:
				// size of 1/4 of an image, smack in the
				// center
void RubberRectangle::init_rectangle(void)
{
  XVTRect image_rect = (XVTBasicWindow&)parent;
  const int width = image_rect.q_width();
  const int height = image_rect.q_height();
  rect.top = height/4;
  rect.left = width/4;
  rect.bottom = rect.top+height/2;
  rect.right = rect.left+width/2;
}

				// This is a real "constructor", that is
				// it is this function that makes the
				// window/rectangle appear
void RubberRectangle::open(void)
{
  assert( !*this );			// Make sure the window isn't here
					// already
  create(parent);			// Create a subwindow based on parent
  show();
}

			// Initialize the Window, return FALSE to cancel
			// This function is called right after the window is
			// created
#if (XVTWS >= X11WS)
static GC Rect_draw_gc;			// Special GC for drawing rect under
					// the X window (only this single GC
					// has to have XOR drawing mode
#endif

BOOLEAN RubberRectangle::initialize(void)
{
  if( ! XVTBasicWindow::initialize() )
    return FALSE;
  CustomCLUT clut;
  clut.set_palette(q_window());
#if (XVTWS >= X11WS)
				// X Window needs special handling again:
				// setting brush to hollow doesn't seem
				// to set the background pixmap to None
  Display * display = (Display *)get_value(SCREEN_WIN,ATTR_X_DISPLAY);
  assert( display != 0 );
  Window x_window = (Window)get_value(q_window(),ATTR_NATIVE_WINDOW);
  XSetWindowBackgroundPixmap(display,x_window,None);
  
  Window root;			// Find kids for a window
  Window parent;
  Window *kids;
  unsigned int nkids;
  assert( XQueryTree(display,x_window,&root,&parent, &kids, &nkids) );
  assert( nkids == 0 );
  XSetWindowBackgroundPixmap(display,parent,None);
  XFree((char *)kids);

				// We need to set XOR mode on the lowest level
				// for X. Note, *don't* set GXxor mode for
				// the "global" GC gotten from
				// get_value(q_window(),
				// 	ATTR_NATIVE_GRAPHIC_CONTEXT);
				// It screws up all the private colormaps
				// Unfortunately, XVT does exactly that
//  GC global_gc = (GC)get_value(q_window(),ATTR_NATIVE_GRAPHIC_CONTEXT);
//  XSetFunction(display,global_gc,GXnoop);
  XGCValues values;
  values.function = GXxor;	// We set XOR mode for only private GC
  values.plane_mask = BlackPixel(display,DefaultScreen(display)) ^ 
    WhitePixel(display,DefaultScreen(display));
  values.foreground = 0xffffffff;
  values.background = 0;
  values.line_style = LineDoubleDash;
  values.line_width = 2;
  Rect_draw_gc = XCreateGC(display,x_window,
			   GCFunction | GCForeground | GCBackground |
			   GCLineWidth | GCLineStyle,&values);
  XClearArea(display,x_window, 0,0,0,0, True);
//  XClearWindow(display,x_window);
#else
  DRAW_CTOOLS draw_tools;
//  CBRUSH brush = { PAT_HOLLOW,  0 };
  ::win_get_draw_ctools(q_window(),&draw_tools);
  draw_tools.mode = M_XOR;
  draw_tools.pen.pat=PAT_RUBBER;
  draw_tools.brush.pat=PAT_HOLLOW;
  win_set_draw_ctools(q_window(),&draw_tools);
#endif

  processed_changed_rect(rect);
  mouse_touchdown_type = MT_none;
  return TRUE;
}

				// Close the window (if it was open)
void RubberRectangle::cancel(void)
{ 
  if( !*this )
    return;				// There is no open window
  XVTBasicWindow::cancel();
}

/*
 *------------------------------------------------------------------------
 *			Drawing and event processing
 */

#if (XVTWS >= X11WS)

				// It is this function that draws the rect
				// We have to use our private GC for drawing
				// So we need to do all the drawing ourselves
void RubberRectangle::draw_immediately(void)
{
  Display * display = (Display *)get_value(SCREEN_WIN,ATTR_X_DISPLAY);
  assert( display != 0 );
  Window x_window = (Window)get_value(q_window(),ATTR_NATIVE_WINDOW);
  XDrawRectangle(display,x_window,Rect_draw_gc,rect.left,rect.top,
		 rect.q_width(),rect.q_height());
}

#else

				// It is this function that draws the rect
void RubberRectangle::draw_immediately(void)
{
  ::win_draw_rect(q_window(),rect);
}

#endif

				// Handle an expose event
void RubberRectangle::draw(void)
{
  draw_immediately();
  parent.draw_immediately();
}

				// It handles mouse pressing events
				// The rest of events are handled by a generic
				// window event handler
				// We ought to handle E_SIZE events here (to
				// scale our window when the parent gets
				// resized). Note, that E_SIZE is also the
				// first event after E_CREATE, always.
				// But we'll do it next time.
void RubberRectangle::handler(EVENT event)
{
					// Handle the mouse movement only
					// when a mouse button kept pressed
  if( event.type == E_MOUSE_DOWN )
  {
    mouse_touchdown_type = classify_pos(mouse_pos = event.v.mouse.where);
  }
  else if( event.type == E_MOUSE_UP )
    mouse_touchdown_type = MT_none;
  else if( event.type == E_MOUSE_MOVE && mouse_touchdown_type != MT_none )
    handle_mouse_move(event.v.mouse.where);
    
  XVTBasicWindow::handler(event);
}


				// Handle the mouse movement from 'mouse_pos'
				// to 'new_point'
void RubberRectangle::handle_mouse_move(const PNT new_point)
{
					// Beware! this win_rect may be
					// (0,0)-(128,128) for a 128x128 window
  XVTRect win_rect = (XVTBasicWindow&)*this;
					// Always make sure we won't drag/
					// stretch rectangle outside of window
  if( !pt_in_rect(win_rect,new_point) )
    return;

  assert( rect.left < rect.right && rect.top < rect.bottom );

  if( mouse_touchdown_type == MT_none )
    return;

  XVTRect new_rect = rect;

  if( mouse_touchdown_type == MT_in_rect )
  {					// Move the rectangle as a whole
    int dx = new_point.h - mouse_pos.h;
    int dy = new_point.v - mouse_pos.v;
    if( new_rect.top + dy < win_rect.top )	// Clip the movement to
      dy = win_rect.top - new_rect.top;		// the window boundaries
    if( new_rect.bottom + dy >= win_rect.bottom )
      dy = win_rect.bottom - new_rect.bottom - 1;
    new_rect.top += dy, new_rect.bottom += dy;
    if( new_rect.left + dx < win_rect.left )
      dx = win_rect.left - new_rect.left;
    if( new_rect.right + dx >= win_rect.right )
      dx = win_rect.right - new_rect.right - 1;
    new_rect.left += dx, new_rect.right += dx;
  }
  else
  {					// Move rectangle boundary(ies)
//    message("rect (%d,%d)-(%d,%d)\n",
//	    rect.left,rect.top,rect.right,rect.bottom);
//    message("mouse was at (%d,%d), now at (%d,%d)\n",
//	    mouse_pos.h,mouse_pos.v,new_point.h,new_point.v);
					// Make sure that rectangle still
					// has at least 1 point width and
					// hight, however squeezed it is
    if( (mouse_touchdown_type & MT_top) &&  new_point.v < new_rect.bottom )
      new_rect.top = new_point.v;

    if( (mouse_touchdown_type & MT_bottom) &&  new_point.v > new_rect.top )
      new_rect.bottom = new_point.v;

    if( (mouse_touchdown_type & MT_left) &&  new_point.h < new_rect.right )
      new_rect.left = new_point.h;

    if( (mouse_touchdown_type & MT_right) &&  new_point.h > new_rect.left )
      new_rect.right = new_point.h;
  }

  draw_immediately();			// Since the draw mode is set to XOR,
					// this immediately gets rid of the
					// old rectangle
  mouse_pos = new_point;

  rect = new_rect;
  assert( rect.left < rect.right && rect.top < rect.bottom );
  assert( rect.left >= win_rect.left && rect.top >= win_rect.top );

  processed_changed_rect(rect);
  draw_immediately();
}

				// Classify the position specify by point
				// relative to 'rect'
				// Note, the mouse position type stays
				// always the same for a held down mouse
				// until the mouse button is released
				// Note, to make it easier to grab a rect
				// boundary, we make it "thicker"
RubberRectangle::MOUSE_POS RubberRectangle::classify_pos(const PNT pt)
{
  const int dx = 1, dy = 1;		// Thickness (uncertainty) of boundary

  if( pt.h < rect.left-dx || pt.h > rect.right+dx )
    return MT_none;			// The point is outside of the rect

  if( pt.v < rect.top-dy || pt.v > rect.bottom+dy )
    return MT_none;

  if( pt.h > rect.left+dx && pt.h < rect.right-dx && pt.v > rect.top+dy &&
      pt.v < rect.bottom-dy )
    return MT_in_rect;

  MOUSE_POS pos = MT_none;

  if( pt.h >= rect.left-dx && pt.h <= rect.left+dx )
    pos = (MOUSE_POS)(pos | MT_left);
  else if( pt.h >= rect.right-dx && pt.h <= rect.right+dx )
    pos = (MOUSE_POS)(pos | MT_right);

  if( pt.v >= rect.top-dy && pt.v <= rect.top+dy )
    pos = (MOUSE_POS)(pos | MT_top);
  else if( pt.v >= rect.bottom-dy && pt.v <= rect.bottom+dy )
    pos = (MOUSE_POS)(pos | MT_bottom);

  return pos;
}
				// Called whenever the rectangle is changed
				// or moved. Return FALSE to kill the 
				// RubberRectangle
BOOLEAN RubberRectangle::processed_changed_rect(const XVTRect& rect) const
{
  return TRUE;
}


